<?php
/**
 * Class to check opening_hours as used by the OpenStreetMap project
 *
 * This code is a port of the javascript code from
 * http://www.netzwolf.info/kartografie/osm/time_domain/
 *
 * PHP version 5.1.0+
 *
 * Copyright (c) 2012, Uwe Steinmann
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *  - Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  - Neither the name of the The PEAR Group nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * @category  OSM
 * @package   OSM_OpeningHours
 * @author    Uwe Steinmann <steinm@php.net>
 * @copyright 2012 Uwe Steinmann
 * @license   http://www.opensource.org/licenses/bsd-license.php New BSD License
 * @version   CVS: $Revision: $
 * @link      http://pear.php.net/package/OSM_OpeningHours
 */

/**
 * Include PEAR package for error handling
 */
require_once 'PEAR.php';

/**
 * Include OSM_OpeningHours_Date class
 */
require_once 'OSM_OpeningHours/Date.php';

/**
 * Include OSM_OpeningHours_Request class
 */
require_once 'OSM_OpeningHours/Request.php';

/**
 * Class for checking opening_hours as used by the OpenStreetMap project
 *
 * @category  OSM
 * @package   OSM_OpeningHours
 * @author    Uwe Steinmann <steinm@php.net>
 * @copyright 2012 Uwe Steinmann
 * @license   http://www.opensource.org/licenses/bsd-license.php New BSD License
 * @link      http://pear.php.net/package/OSM_OpeningHours
 */
class OSM_OpeningHours {

	//----------------------------------------------------------------------------
	//	initializing arrays
	//----------------------------------------------------------------------------
	public static $seasons = array(
			'summer' => 2, 'winter' => 4
		);

	public static $months = array(
			'jan'=> 1, 'feb'=> 2, 'mar'=> 3, 'apr'=> 4, 'may'=> 5, 'jun'=> 6,
			'jul'=> 7, 'aug'=> 8, 'sep'=> 9, 'oct'=>10, 'nov'=>11, 'dec'=>12
		);

	public static $weekdays = array(
			'su'=>0, 'mo'=>1, 'tu'=>2, 'we'=>3, 'th'=>4, 'fr'=>5, 'sa'=>6
		);

	public static $movables = array(
			'easter'=>'easter'
		);

	public static $holidays = array(
			'sh'=>'school', 'ph'=>'public'
		);

	public static $events = array(
			'sunrise' => 'sunrise', 'sunset' => 'sunset'
		);

	public static $holidayDates = array();

	public static $defaultRegion = 'de';

	/**
	 * This function is a replacement for the javascript function
	 */
	static function parseInt($str, $base=10) {
		$i = intval($str, $base);
		if($i === 0 && !ctype_digit($str))
			return null;
		else
			return $i;
	}

	/**
	 * This function is a replacement for the javascript function
	 */
	static function isNaN($nr) {
		return !is_numeric($nr);
	}

	//----------------------------------------------------------------------------
	// some commonly used words which should not be used
	//----------------------------------------------------------------------------

	public static $wrong_words = array( /* {{{ */
		'Please ommit "<ko>".' => array(
			'h'=>	''
		),
		'Please use notation "<ok>" for "<ko>".' => array(
			'to'=>	'-'
		),
		'Bitte verzichte auf "<ko>".' => array(
			'uhr'=>	''
		),
		'Bitte benutze die Schreibweise "<ok>" für "<ko>".' => array(
			'und'=>	','
		),
		'Bitte benutze die englische Abkürzung "<ok>" für "<ko>".' => array(
			'feiertag'=>'PH',
			'feiertage'=>'PH',
			'feiertagen'=>'PH'
		),
		'S\'il vous plaît utiliser "<ok>" pour "<ko>".' => array(
			'fermé'=>'off',
			'et'=>	',',
			'à'=>	'-'
		),
		'Neem de engelse afkorting "<ok>" voor "<ko>" alstublieft.' => array(
			'feestdag'=>'PH',
			'feestdagen'=>'PH'
		)); /* }}} */

	public static $wrong_seasons = array( /* {{{ */
		'Bitte benutze die englische Schreibweise "<ok>" für "<ko>".' => array(
			'sommer'=>2
		)); /* }}} */

	public static $wrong_months = array( /* {{{ */
		'Please use the englisch abbreviation "<ok>" for "<ko>".' => array(
			'january'=>1,
			'february'=>2,
			'march'=>	3,
			'april'=>	4,
			'may'=>	5,
			'june'=>	6,
			'july'=>	7,
			'august'=>	8,
			'september'=>9,
			'october'=>10,
			'november'=>11,
			'december'=>12
		),
		'Bitte benutze die englische Abkürzung "<ok>" für "<ko>".' => array(
			'januar'=>	1,
			'februar'=>2,
			'märz'=>	3,
			'maerz'=>	3,
			'mai'=>	5,
			'juni'=>	6,
			'juli'=>	7,
			'okt'=>	10,
			'oktober'=>10,
			'dez'=>	12,
			'dezember'=>12
		),
		'S\'il vous plaît utiliser l\'abréviation "<ok>" pour "<ko>".' => array(
			'janvier'=>1,
			'février'=>2,
			'fév'=>	2,
			'mars'=>	3,
			'avril'=>	4,
			'avr'=>	4,
			'mai'=>	5,
			'juin'=>	6,
			'juillet'=>7,
			'août'=>	8,
			'aoû'=>	8,
			'septembre'=>9,
			'octobre'=>10,
			'novembre'=>11,
			'décembre'=>12
		),
		'Neem de engelse afkorting "<ok>" voor "<ko>" alstublieft.' => array(
			'januari'=>1,
			'februari'=>2,
			'maart'=>	3,
			'mei'=>	5,
			'augustus'=>8
		)); /* }}} */

	public static $wrong_weekdays = array( /* {{{ */
		'Please use the abbreviation "<ok>" for "<ko>".' => array(
			'sun'=>	0,
			'sunday'=>	0,
			'mon'=>	1,
			'monday'=>	1,
			'tue'=>	2,
			'tuesday'=>2,
			'wed'=>	3,
			'wednesday'=>3,
			'thu'=>	4,
			'thursday'=>4,
			'fri'=>	5,
			'friday'=>	5,
			'sat'=>	6,
			'saturday'=>6
		),
		'Bitte benutze die englische Abkürzung "<ok>" für "<ko>".' => array(
			'so'=>	0,
			'son'=>	0,
			'sonntag'=>0,
			'montag'=>	1,
			'di'=>	2,
			'die'=>	2,
			'dienstag'=>2,
			'mi'=>	3,
			'mit'=>	3,
			'mittwoch'=>3,
			'do'=>	4,
			'don'=>	4,
			'donnerstag'=>4,
			'fre'=>	5,
			'freitag'=>5,
			'sam'=>	6,
			'samstag'=>6
		),
		'S\'il vous plaît utiliser l\'abréviation "<ok>" pour "<ko>".' => array(
			'dim'=>	0,
			'dimanche'=>0,
			'lu'=>	1,
			'lun'=>	1,
			'lundi'=>	1,
			'mardi'=>	2,
			'mer'=>	3,
			'mercredi'=>3,
			'je'=>	4,
			'jeu'=>	4,
			'jeudi'=>	4,
			've'=>	5,
			'ven'=>	5,
			'vendredi'=>5,
			'samedi'=>	6
		),
		'Neem de engelse afkorting "<ok>" voor "<ko>" alstublieft.' => array(
			'zo'=>	0,
			'zon'=>	0,
			'zontag'=>	0,
			'maandag'=>1,
			'din'=>	2,
			'dinsdag'=>2,
			'wo'=>	3,
			'woe'=>	3,
			'woensdag'=>3,
			'donderdag'=>4,
			'vr'=>	5,
			'vri'=>	5,
			'vrijdag'=>5,
			'za'=>	6,
			'zat'=>	6,
			'zaterdag'=>6
		)); /* }}} */

	/**
	 * Compare function for sorting arrays
	 */
	static function cmp($a, $b) { return $a['t']-$b['t']; }

	/**
	 * evaluate "opening_hours"-expression
	 * for a certain time and a certain region
	 *
	 * @param string $expression opening_hours
	 * @param array/object either a object of class OSM_OpeningHours_Date or
	 *        an array with the date/time components.
	 * @param array $option
	 * @return object OSM_OpeningHours_Request
	 */
	function evaluateInTime($expression, $checktime = null, $option = array()) { /* {{{ */

		$request = $this->createTimestamp ($checktime, null, 0);
		$mintime = new OSM_OpeningHours_Date($request->y, $request->m-1, $request->d, $request->H, $request->M, $request->S);
		$this->scan ($request, $expression);
		$request->option = $option;
		$this->evaluateRequest ($request);

		//--------------------------------------------------------------
		//	no match → try daywrap
		//--------------------------------------------------------------

		if ($request->value!=true) {
			$r24 = $this->createTimestamp ($checktime, null, 24);
			$r24->parsedExpression = $request->parsedExpression;
			$r24->option = $option;
			$this->evaluateRequest ($r24);

			$times = array_merge($r24->times, $request->times);

			if ($r24->usedrule) $r24->usedrule .= ' (with daywrap)';
			if ($r24->value==true || $r24->value==null && $r24->comment!=null && $request->comment==null) $request=$r24;
			$request->times = $times;
		}

		for ($i=0; $i<count($request->times); $i++) {
			if ($request->times[$i]['t']<$mintime) array_splice($request->times, $i--,1);
		}

		return $request;
	} /* }}} */

	/**
	 * evaluate "collection_time"-expression
	 * for a certain time and a certain region
	 */
	function evaluateNextTime($expression, $checktime = null, $option = array()) { /* {{{ */

		$request = $this->createTimestamp ($checktime, null, 0);
		$this->scan ($request, $expression);
		$request->times = array();

		$mintime = new OSM_OpeningHours_Date($request->y, $request->m-1, $request->d, $request->H, $request->M, $request->S);
		//--------------------------------------------------------------
		//	loop
		//--------------------------------------------------------------

		for ($offset=-1; $offset<7; $offset++) {

			$check = $this->createTimestamp(new OSM_OpeningHours_Date(
					$request->y,
					$request->m-1,
					$request->d+$offset,
					12));

			$check->parsedExpression = $request->parsedExpression;
			$check->option = $option;

			$this->evaluateRequest ($check);

			//--------------------------------------------------------------
			//	check and append to result
			//--------------------------------------------------------------

			foreach($check->times as $tmp) {
				if($tmp['t']->getTime() > $mintime->getTime())
					$request->times[] = $tmp;
			}

			$errors	= $check->errors;
			$messages= $check->messages;

			if (count($request->times)>=1) break;
		}

		$request->errors = array_merge($request->errors, $errors);
		$request->messages = array_merge($request->messages, $messages);

		return $request;
	} /* }}} */

	//--------------------------------------------------------------
	//	error handler
	//--------------------------------------------------------------

	function getRuleElements ($request, $start) { /* {{{ */
		$out = array();
		for ($i=$start; isset($request->parsedExpression[$i]['t']); $i++) {
			$out[] = $request->parsedExpression[$i]['s'];
		}
		return $out;
	} /* }}} */

	function log($request, $msg, &$list) { /* {{{ */
		$out = $this->getRuleElements ($request, $request->rule_start);
		$pos = count($request->parsedExpression) - count($request->tokens) - $request->rule_start;
		array_splice($out, max(0, min(count($out), $pos)), 0, array('««»»'));
		$list[] =  $msg . ' at "' . implode(' ', $out) . '"';
	} /* }}} */

	function scanError($request, $msg) { /* {{{ */
		$request->errors[] = $msg;
	} /* }}} */

	function error($request, $msg) { /* {{{ */
		if (!$request->errors) $request->errors= array();
		$this->log ($request, $msg, $request->errors);
	} /* }}} */

	function warning($request, $msg) { /* {{{ */
		$this->log ($request, $msg, $request->warnings);
	} /* }}} */

	//--------------------------------------------------------------
	//	evaluate
	//--------------------------------------------------------------

	function evaluateRequest($request) { /* {{{ */

		$request->range_match = true;
		$request->value	= false;
		$request->comment	= null;
		$request->start	= null;
		$request->tokens	= $request->parsedExpression; //.slice();

		while (count($request->tokens)>=1) {

			if (isset($request->tokens[0]['s']) && $request->tokens[0]['s']=='||') {
				array_shift($request->tokens);
				if ($request->value) $request->skip = true;
				if ($request->skip) continue;
				$request->range_match = true;
				$request->value	= false;
				$request->comment	= null;
				continue;
			}

			if (!isset($request->tokens[0]['t'])) {
				array_shift($request->tokens);
				continue;
			}

			//--------------------------------------------------------------
			//	evaluate a single rule
			//--------------------------------------------------------------

			$request->rule_start = count($request->parsedExpression)-count($request->tokens);

			$error = $this->evaluateBasicRule ($request);

			//--------------------------------------------------------------
			//	error handler
			//--------------------------------------------------------------

			if (gettype($error)=='string' || isset($request->tokens[0]['t'])) {

				if (!$error) $error = 'Syntax error';

				$this->error ($request, $error);

				while (isset($request->tokens[0]['t']) && $request->tokens[0]['t']) array_shift($request->tokens);

				if ($request->skip) continue;

				$request->value	= null;
				$request->comment	= 'invalid specification';
				$request->start	= $request->rule_start;

				continue;
			}

			//--------------------------------------------------------------
			//	skip
			//--------------------------------------------------------------

			if ($request->skip) continue;

			//--------------------------------------------------------------
			//	applicable
			//--------------------------------------------------------------

			if (!$request->range_match || !$request->day_match) continue;

			//--------------------------------------------------------------
			//	quick fix for informal dayrange specifications
			//--------------------------------------------------------------

			if (gettype($request->range_match)=='string') {

				if ($request->rule_value) $request->rule_value = null;

				$request->rule_comment = $request->rule_comment ?
					$request->range_match + " " + $request->rule_comment :
					$request->range_match;
			}

			//--------------------------------------------------------------
			//	store value
			//--------------------------------------------------------------

			$request->value	= $request->rule_value;
			$request->comment	= $request->rule_comment;
			$request->start	= $request->rule_start;
			$request->minutes	= $request->rule_minutes;
		}

		if ($request->start !== null)
			$request->usedrule = implode(' ', $this->getRuleElements($request, $request->start));

		//------------------------------------------------------------
		//	process minutes
		//------------------------------------------------------------

		$minutes = array();

		foreach ($request->minutes as $entry) {
			if ($entry['v']==false) {
				for ($i=0; $i<count($minutes); $i++) {
					if ($entry['f']<=$minutes[$i]['t'] && $minutes[$i]['t']<=$entry['l'])
						array_splice($minutes, $i--, 1);
				}
				continue;
			}
			if ($entry['s']>=1) {
				for ($t=$entry['f']; $t<=$entry['l']; $t+=$entry['s']) {
					$minutes[] = array('t' => $t, 'c' => $entry['c']);
				}
				continue;
			}
			$minutes[] = array('t' => $entry['f'], 'c' => $entry['c']);
		}

		usort($minutes, 'OSM_OpeningHours::cmp');

		//--------------------------------------------------------------
		//	transform to times
		//--------------------------------------------------------------

		$tmpdate = new OSM_OpeningHours_Date ($request->y, $request->m-1, $request->d, 12, 0, 0);
		$base = $tmpdate->getTime() - 12 *3600*1000;

		$request->times = array();
		foreach($minutes as $tmp) {
			$request->times[] =  array('t' => new OSM_OpeningHours_Date ($base + $tmp['t']*60*1000), 'c' => $tmp['c']);
		}
	} /* }}} */

	//----------------------------------------------------------------------------
	//	evaluate single rule:
	//	string:	error
	//	null:	no error
	//----------------------------------------------------------------------------

	function evaluateBasicRule($request) { /* {{{ */

		//------------------------------------------------------------
		//	init
		//------------------------------------------------------------

		$request->day_match	= false;
		$request->time_match	= false;
		$request->rule_value	= false;
		$request->rule_comment	= null;
		$request->rule_minutes	= array();

		//------------------------------------------------------------
		//	rule_sequence - calendar_ranges
		//------------------------------------------------------------

		$error = $this->evaluateCalendarRanges ($request);

		if (gettype($error)=='string') return $error;

		//------------------------------------------------------------
		//	Empty rule
		//------------------------------------------------------------

		if (!isset($request->tokens[0]['t'])) return;

		//------------------------------------------------------------
		//	Basic Rule
		//------------------------------------------------------------

		for (;;) {

			//------------------------------------------------------------
			//	Times for Days
			//------------------------------------------------------------

			$error = $this->evaluateTimesForDays ($request);

			if (gettype($error)=='string') return $error;

			//-------------------------------------------------
			//	{ "," Times for Days }
			//-------------------------------------------------

			if ($this->C($request->tokens, ',')) continue;

			//-------------------------------------------------
			//	error correction
			//-------------------------------------------------

			if(isset($request->tokens[0]['t'])) {
				switch ($request->tokens[0]['t']) {
				case 'week':	// daylist:	week
				case 'd':	// daylist:	mday
				case 'h':	// daylist:	holiday
				case 'w':	// daylist:	weekday
					$this->error ($request, 'Comma expected');
					continue;
				}
			}

			return;
		}
	} /* }}} */

	//----------------------------------------------------------------------------
	//	evaluate single Days+Times group
	//	string:	if error
	//	null:	if no error
	//----------------------------------------------------------------------------

	function evaluateTimesForDays($request) { /* {{{ */

		//------------------------------------------------------------
		//	24/7
		//------------------------------------------------------------

		if ($this->C($request->tokens,'24/7')) {
			if ($request->H>=24) return;
			$request->day_match	= true;
			$request->time_match	= true;
			$request->rule_value	= true;
			return;
		}

		//------------------------------------------------------------
		//	open
		//------------------------------------------------------------

		if ($this->C($request->tokens,'open')) {
			$comment = $this->CG($request->tokens, '"');
			if ($request->H>=24) return;
			$request->day_match	= true;
			$request->time_match	= true;
			$request->rule_value	= true;
			$request->rule_comment 	= $comment;
			return;
		}

		//------------------------------------------------------------
		//	closed
		//------------------------------------------------------------

		if ($this->C($request->tokens,'closed') || $this->C($request->tokens,'off')) {
			$comment = $this->CG($request->tokens, '"');
			$request->day_match	= true;
			$request->time_match	= true;
			$request->rule_value	= false;
			$request->rule_comment 	= $comment;
			return;
		}

		//------------------------------------------------------------
		//	unknown / "..."
		//------------------------------------------------------------

		if ($this->C($request->tokens,'unknown') || $request->tokens[0]['t']=='"') {
			$comment = $this->CG($request->tokens, '"');
			$request->day_match	= true;
			$request->time_match	= true;
			$request->rule_value	= null;
			$request->rule_comment 	= $comment;
			return;
		}

		//------------------------------------------------------------
		//	<day_list>
		//		string:	error
		//		null:	no daylist
		//		true:	in daylist
		//		false:	not in daylist
		//------------------------------------------------------------

		$day_match = $this->evaluateDayList ($request);

		if (gettype($day_match)=='string') return $day_match;

		if ($day_match!==false) $request->day_match = true;

		//------------------------------------------------------------
		//	<day_list> "off"
		//------------------------------------------------------------

		if ($day_match !== null &&
			($this->C($request->tokens, 'off')||$this->C($request->tokens, 'closed'))) {
			$comment = $this->CG($request->tokens, '"');
			$request->time_match	= true;
			$request->rule_value	= false;
			$request->rule_comment	= $comment;
			return;
		}

		//------------------------------------------------------------
		//	Times
		//------------------------------------------------------------

		$in_times = $this->evaluateTimes ($request, $day_match!==null);

		if ($day_match===null) $day_match = true;

		if (gettype($in_times)=='string') return $in_times;

		//------------------------------------------------------------
		//	Status	(open, closed, unknown)
		//------------------------------------------------------------

		if ($this->C($request->tokens, 'open')) {
			$value	= true;
			$comment	= $this->CG($request->tokens, '"');
		} else if ($this->C($request->tokens, 'closed')||$this->C($request->tokens, 'off')) {
			$value	= false;
			$comment	= $this->CG($request->tokens, '"');
		} else if ($this->C($request->tokens, 'unknown')) {
			$value	= null;
			$comment	= $this->CG($request->tokens, '"');
		} else if (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='"') {
			$value	= null;
			$comment	= $this->G($request->tokens);
		} else {
			$value	= true;
			$comment = null;
		}

		//------------------------------------------------------------
		//	If daymatch, save times for "NextTime"
		//------------------------------------------------------------

		if ($day_match==true) foreach ($request->_times as $obj) {
			$obj['v'] = $value;
			$obj['c'] = $comment;
			$request->rule_minutes[] = $obj;
		}

		//------------------------------------------------------------
		//	Rule definitely not applicable
		//------------------------------------------------------------

		if ($day_match==false || $in_times==false) return;

		//------------------------------------------------------------
		//	Rule definitely applicable
		//------------------------------------------------------------

		if ($day_match==true && $in_times==true) {
			$request->rule_value	= $value;
			$request->rule_comment	= $comment;
			$request->time_match	= true;
			return;
		}

		//------------------------------------------------------------
		//	Errorhandling shortcut
		//------------------------------------------------------------

		if (gettype($request->rule_value)=='boolean' && $request->rule_value==value) return;

		//------------------------------------------------------------
		//	Something went wong
		//------------------------------------------------------------

		$request->rule_value = null;
		$request->rule_comment  = null;
		$request->time_match = true;
		return;
	} /* }}} */

	//============================================================================
	//
	//	CALENDAR RANGES
	//
	//============================================================================

	//----------------------------------------------------------------------------
	//	evaluate ranges:
	//		r.new_range	- range provided?
	//		r.range_match	- $request matches range provided?
	//	returns:
	//		string:	error
	//		true:	range provided
	//		false:	no range provided
	//----------------------------------------------------------------------------

	function evaluateCalendarRanges($request) { /* {{{ */

		//--------------------------------------------------------
		//	[ calendar_ranges ]
		//--------------------------------------------------------

		switch ($request->tokens[0]['t']) {
		case '"':
			if ($request->tokens[1]['t']!=':') return false;
		case 'y':
		case 'm':
		case 'v':
		case 's':
			break;
		default:
			return false;
		}

		//--------------------------------------------------------
		//	calendar_ranges
		//--------------------------------------------------------

		$request->range_match  = false;

		//--------------------------------------------------------
		//	sticky values
		//--------------------------------------------------------

		$request->thisMonth = null;
		$request->thisYear  = $request->y;
		$request->yearSet   = false;

		//--------------------------------------------------------
		//	list of CalendarDays
		//--------------------------------------------------------

		for (;;) {

			//-----------------------------------------
			//	year
			//-----------------------------------------

			if ($request->tokens[0]['t']=='y') {
				$request->thisYear  = $this->G($request->tokens);
				$request->yearSet   = true;
				$request->thisMonth = null;
			}

			//--------------------------------------------------------
			//	month [ - month ]
			//--------------------------------------------------------

			if (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='m' && (!isset($request->tokens[1]['t']) || $request->tokens[1]['t']!='n') && (!isset($request->tokens[1]['t']) || $request->tokens[1]['t']!='w')) {

				$first = $this->G($request->tokens);
				$last  = $first;

				if ($this->C($request->tokens, '-')) {
					if ($request->tokens[0]['t']!='m') return 'Month expected';
					$last = $this->G($request->tokens);
				}

				if ($first <= $request->m && $request->m <= $last && $request->y==$request->thisYear)
					$request->range_match=true;

				if ($first>$last && !$request->yearSet && ($request->m>=$first || $request->m<=$last))
					$request->range_match=true;

				if ($first>$last && $request->yearSet) {
					if ($request->y==$request->thisYear   && $request->m>=$first) $request->range_match=true;
					if ($request->y==$request->thisYear+1 && $request->m<=$last ) $request->range_match=true;
				}

			} else if ($request->tokens[0]['t']=='s') {

				$season = $this->G($request->tokens);

				if (!$request->yearSet || $request->thisYear==$request->y) {

					switch ($season) {
					case 2:
						$this->warning ($request, 'Assuming "May-Oct" for "summer"');
						if ($request->m>=5 && $request->m<11) $request->range_match=true;
						break;
					case 4:
						$this->warning ($request, 'Assuming "Nov-Apr" for "winter"');
						if ($request->m<=4 || $request->m>10) $request->range_match=true;
						break;
					}
				}

			} else if ($request->tokens[0]['t']=='"') {

				//------------------------------------------------
				//	textdescription
				//------------------------------------------------

				$comment = $this->G ($request->tokens);
				$this->warning ($request, 'Description "'.comment.'" is not evaluated.');
				$request->range_match = $comment;

			} else {

				//------------------------------------------------
				//	date_with_offset
				//------------------------------------------------

				$first = $this->evaluateDateWithOffset ($request);
				if (gettype($first) == 'string') return $first;

				//------------------------------------------------
				//	[ - date_with_offset ]
				//------------------------------------------------

				$last = $first;
				if ($this->C($request->tokens,'-')) {
					$last = $this->evaluateDateWithOffset ($request);
					if (gettype($last) == 'string') return $last;
				} else if ($this->C($request->tokens,'+')) {
					$last = 999999999;
				}

				//------------------------------------------------
				//	check if date in interval
				//	dates may be null
				//------------------------------------------------

				if ($first!=null && $last!=null) {

					if ($first <= $request->e && $request->e <= $last)
						$request->range_match=true;

					if ($last<$first && !$request->yearSet && ($request->e>=$first || $request->e<=$last))
						$request->range_match=true;

					if ($last<$first && $request->yearSet)
						$this->warning ($request, 'Empty date range');

				} else if (!$request->range_match) {

					$request->range_match = null;

				}
			}

			//--------------------------------------------------------
			//	"," starts next range
			//--------------------------------------------------------

			if (!$this->C($request->tokens, ',')) break;
		}

		if (!$this->C($request->tokens, ':'))
			$this->warning ($request, 'Missing ":" after range');

		return true;
	} /* }}} */

	//----------------------------------------------------------------------------
	//	evaluate date_with_offset
	//	string:	any error
	//	null:	unknown date
	//	number:	epoch value of given day
	//----------------------------------------------------------------------------

	function evaluateDateWithOffset($request) { /* {{{ */

		$day = $this->evaluateDate ($request);

		if (gettype($day) == 'string') return $day;

		//------------------------------------------------------
		//	wday_offset
		//------------------------------------------------------

		if (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='+' && isset($request->tokens[1]['w']) && $request->tokens[1]['t']=='w') {
			if ($day!=null) $day += 7 - ($this->W($day)-$request->tokens[1]['v']+7)%7;
			array_splice($request->tokens, 0, 2);
		} elseif (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='-' && isset($request->tokens[1]['w']) && $request->tokens[1]['t']=='w') {
			if ($day!=null) $day -= 7 - ($request->tokens[1]['v']-$this->W($day)+7)%7;
			array_splice($request->tokens, 0, 2);
		}

		//------------------------------------------------------
		//	± <number> $days
		//------------------------------------------------------

		if (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='+' && isset($request->tokens[1]['t']) && $request->tokens[1]['t']=='n' && isset($request->tokens[2]['t']) && $request->tokens[2]['t']=='days') {
			if ($day!=null) $day += $request->tokens[1]['v'];
			array_splice($request->tokens,0,3);
		} elseif (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='-' && isset($request->tokens[1]['t']) && $request->tokens[1]['t']=='n' && isset($request->tokens[2]['t']) && $request->tokens[2]['t']=='days') {
			if ($day!=null) $day -= $request->tokens[1]['v'];
			array_splice($request->tokens,0,3);
		}

		return $day;
	} /* }}} */

	//----------------------------------------------------------------------------
	//	evaluate date
	//	string:	any error
	//	null:	unknown date
	//	number:	epoch value of given day
	//----------------------------------------------------------------------------

	function date2day($request, $mday) { /* {{{ */

		if ($request->thisMonth==2 && $mday==29 && !$request->yearSet)
			return "Feb 29 only allowed if year specified";

		$day  = $this->E ($request->thisYear, $request->thisMonth, $mday);

		if ($mday>=1 && $day<=$this->E ($request->thisYear, $request->thisMonth+1, 0))
			return $day;

		$this->error ($request, 'Invalid day of month');
		return null;
	} /* }}} */

	function evaluateDate($request) { /* {{{ */

		//-----------------------------------------
		//	day without month
		//-----------------------------------------

		if ($request->tokens[0]['t']=='n' && $request->thisMonth) {
			return $this->date2day ($request, $this->G($request->tokens));
		}

		//-----------------------------------------
		//	year
		//-----------------------------------------

		if ($request->tokens[0]['t']=='y') {
			$request->thisYear  = $this->G($request->tokens);
			$request->yearSet   = true;
			$request->thisMonth = null;
		}

		//-----------------------------------------
		//	variable date
		//-----------------------------------------

		if ($request->tokens[0]['t']=='v')
			return $this->evaluateMovable($request->thisYear, $this->G($request->tokens));

		//-----------------------------------------
		//	month
		//-----------------------------------------

		if ($request->tokens[0]['t']!='m') return 'Month expected';
		$request->thisMonth = $this->G($request->tokens);

		//-----------------------------------------
		//	mday or wday ?
		//-----------------------------------------

		switch ($request->tokens[0]['t']) {
		case 'n':
			return $this->date2day ($request, $this->G($request->tokens));
		case 'w':
			$wday=$this->G($request->tokens);
			if (!$this->C($request->tokens, '[')) return '"[" expected';

			$d;
			if ($this->C($request->tokens,'-')) {
				if ($request->tokens[0]['t'] != 'n') return "Weeknumber 1-5 expected";
				$d = $this->E($request->thisYear, $request->thisMonth+1, 1-7*$this->G($request->tokens));
			} else {
				if ($request->tokens[0]['t'] != 'n') return "Weeknumber 1-5 expected";
				$d = $this->E($request->thisYear, $request->thisMonth, 7*$this->G($request->tokens)-6);
			}
			if (!$this->C($request->tokens, ']')) return '"]" expected';
			return $d + ($wday - $this->W(d) + 7) % 7;
		}

		return 'Weekday or day of month expected';
	} /* }}} */

	//----------------------------------------------------------------------------
	//	evaluate movable days
	//	null:	any error
	//	number:	epoch for this day
	//----------------------------------------------------------------------------

	function evaluateMovable($year, $name) { /* {{{ */

		switch ($name) {
		case 'easter':
			$a = $year % 19;
			$b = $year % 4;
			$c = $year % 7;
			$k = floor ($year / 100);
			$p = floor ((8*$k + 13) / 25);
			$q = floor ($k / 4);
			$M = (15 + $k - $p - $q) % 30;
			$N = (4 + $k - $q) % 7;
			$d = (19*$a + $M) % 30;
			$e = (2*$b + 4*$c + 6*$d + $N) % 7;
			$d = (22 + $d + $e);
			return $this->E ($year, 3, $d<57 ? $d : $d-7);
		}

		$this->error ($request, 'No date for movable "'.$name.'" in year "'.$year.'"');
		return null;
	} /* }}} */

	//============================================================================
	//
	//	DAY LIST  incl.  WEEKS
	//
	//============================================================================

	//----------------------------------------------------------------------------
	//	string:	error
	//	null:	empty
	//	true:	not empty, match
	//	false:	not empty, no match
	//----------------------------------------------------------------------------

	function evaluateDayList($request) { /* {{{ */

		switch ($request->tokens[0]['t']) {
		case 'd':
		case 'w':
		case 'h':
		case 'week':
			break;
		default:
			return null;
		}

		//------------------------------------------
		//	Days
		//------------------------------------------

		$result = false;

		for (;;) {

			$inweek = true;

			//------------------------------------------
			//	Weeks n [ - n [ / n ]]
			//------------------------------------------

			if ($this->C($request->tokens, 'week')) {

				$inweek = false;

				do {
					if ($request->tokens[0]['t']!='n') return 'Weeknumber expected';

					$first = $this->G($request->tokens);
					$last  = $first;

					if ($this->C($request->tokens, '-')) {

						//-------------------------------------
						//	week first - last
						//-------------------------------------

						if ($request->tokens[0]['t']!='n') return 'Weeknumber expected';
						$last = $this->G($request->tokens);

						//-------------------------------------
						//	week first - last / step
						//-------------------------------------

						if ($this->C($request->tokens, '/')) {
							if ($request->tokens[0]['t']!='n') return 'Number expected';
							$step=$this->G($request->tokens);
						}
					}

					if (!isset($step)) $step=1;

					if ($first <= $request->yw && $request->yw <= $last &&
						($request->yw-$first)%$step==0) $inweek = true;
				} while ($this->C($request->tokens, ','));
			}

			//------------------------------------------
			//	Holidays-Prefix
			//------------------------------------------

			$cond = $inweek;

			while (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='h' &&
				isset($request->tokens[1]['t']) && ($request->tokens[1]['t']=='d'||$request->tokens[1]['t']=='w'||$request->tokens[1]['t']=='h')) {
				if (!$this->evaluateHoliday ($request, $this->G($request->tokens))) $cond=false;
			}

			//------------------------------------------
			//	Monthday or Weekday or Holiday
			//------------------------------------------

			if(!isset($request->tokens[0]['t'])) {
				return 'Day of month, weekday or type of holiday expected';
			} else {
				switch ($request->tokens[0]['t']) {

				//------------------------------------------
				//	Monthdays
				//------------------------------------------

				case 'd':
					//------------------------------------------------
					//	mday | mday - mday
					//------------------------------------------------

					$first = $this->G($request->tokens);
					$last = $first;
					$step;

					if ($this->C($request->tokens, '-')) {
						if ($request->tokens[0]['t']!='d') return 'Day of month expected';
						$last = $this->G($request->tokens);
						//-------------------------------------
						//	/<step> (for odd/even))
						//-------------------------------------
						if ($this->C($request->tokens, '/')) {
							if ($request->tokens[0]['t']!='n') return 'Number expected';
							$step=$this->G($request->tokens);
						}
					}
					if (!isset($step)) $step=1;
					if ($first <= $request->d && $request->d <= $last &&
							($request->d-$first)%$step==0 && $cond)
						$result = true;
					break;

				//------------------------------------------
				//	Weekdays
				//------------------------------------------

				case 'w':
					$weekday = $this->G($request->tokens);

					// Check if the next token starts a weekday range
					if(!isset($request->tokens[0]['t'])) {
						if ($request->w==$weekday && $cond)
							$result = true;
					} else {
						switch ($request->tokens[0]['t']) {

						//----------------------------------
						//	wday
						//----------------------------------

						default:
							if ($request->w==$weekday && $cond)
								$result = true;
							break;

						//----------------------------------
						//	wday - wday
						//----------------------------------

						case '-':
							array_shift($request->tokens);
							if ($request->tokens[0]['t'] != 'w') return null;
							if (($request->w-$weekday+7)%7 <= ($this->G($request->tokens)-$weekday+7)%7
									&& $cond)
								$result = true;
							break;

						//----------------------------------
						//	wday [...]
						//----------------------------------

						case '[':
							array_shift($request->tokens);
							$found=false;
							for (;;) {
								switch ($request->tokens[0]['t']) {
								default:
									return 'Number expected';
								case 'n':
									$snth = $this->G($request->tokens);
									$enth = $snth;
									if ($this->C($request->tokens,'-')) {
										if ($request->tokens[0]['t']!='n')
											return 'Number expected';
										$enth = $this->G($request->tokens);
									}
									if ($snth*7-6 <= $request->d && $request->d <= $enth*7)
										$found = true;
									break;

								case '-':
									array_shift($request->tokens);
									if ($request->tokens[0]['t'] != 'n')
										return 'Number expected';
									$start = $request->l - $this->G($request->tokens)*7;
									if ($start <= $request->d && $request->d < $start+7)
										$found = true;
								}

								if (!$this->C($request->tokens, ',')) break;
							}

							if (!$this->C($request->tokens, ']')) return '"]" expected';

							if ($found && $request->w==$weekday && $cond)
								$result = true;
						}
					}

					break;

				//------------------------------------------
				//	Holidays
				//------------------------------------------

				case 'h':
					$type  = $this->G($request->tokens);
					$offset= 0;
					if (isset($request->tokens[0]['t']) && ($request->tokens[0]['t']=='+' || $request->tokens[0]['t']=='-') && isset($request->tokens[1]['t']) && $request->tokens[1]['t']=='n' && isset($request->tokens[2]['t']) && $request->tokens[2]['t']=='days') {
						$offset = ($request->tokens[0]['t']=='-' ? -1 : +1) * $request->tokens[1]['v'];
						array_splice($request->tokens,0,3);
					}
					$r = $this->evaluateHoliday ($request, $type, $offset);
					if (!$result && $cond) $result = $r;
					break;

				default:
					return 'Day of month, weekday or type of holiday expected';
				}
			}

			//--------------------------------------------------------
			//	"," starts next selection
			//--------------------------------------------------------

			if (!$this->C ($request->tokens, ',')) return $result;
		}
	} /* }}} */

	//----------------------------------------------------------------------------
	//	true:	parameter day is holiday
	//	false:	parameter day is not holiday
	//	null:	unknown
	//----------------------------------------------------------------------------

	function evaluateHoliday($request, $type, $offset) { /* {{{ */

		if (!isset($request->option['region'])) {
			$this->warning ($request, 'Assuming "'.self::$defaultRegion.'" for region');
			$request->option['region'] = self::$defaultRegion;
		}

		$region = strtolower($request->option['region']);

		$holidayfunc = 'holidays_'.$region;
		if (!method_exists($this, $holidayfunc)) {
			$this->warning ($request,'No holiday list for type "'.$type.'" and region "'.$region.'"');
			return false;
		}

		if (!isset($this->holidayDates[$region])) $this->holidayDates[$region]=array();
		if (!isset($this->holidayDates[$region][$type])) $this->holidayDates[$region][$type]=array();

		$date  = $request->e - $offset;
		$years = array($request->y-1, $request->y, $request->y+1);

		/* TODO I can see why it is needed to calculate the holidays
		 * for last and next year
		 */
		$years = array($request->y);
		foreach ($years as $year) {
			if(!isset($this->holidayDates[$region][$type][$year]))
				if(!$this->holidayDates[$region][$type][$year] = $this->$holidayfunc($year, $type)) {
				$this->warning ($request,
					'Holiday type "'.$type.'" for region "'.$region.'" not configured');
				return false;
			}

			$list = $this->holidayDates[$region][$type][$year];
			foreach ($list as $holidayName=>$entry) {
				if ($list[$holidayName]!=$date) continue;
				$this->warning ($request,
					'Found "'.$holidayName.'" for holiday type "'.$type.'"');
				return true;
			}
		}
		return false;
	} /* }}} */

	function initHolidays($region, $type, $year) { /* {{{ */
		$holidayfunc = 'holidays_'.$region;
		if(!isset($this->holidayDates[$region][$type][$year]))
			if(!$this->holidayDates[$region][$type][$year] = $this->$holidayfunc($year, $type)) {
			/*
			$this->warning ($request,
				'Holiday type "'.$type.'" for region "'.$region.'" not configured');
			*/
			return false;
		}
	} /* }}} */

	function addHoliday($region, $type, $name, $year, $month, $day) { /* {{{ */
		if(!isset($this->holidayDates[$region][$type][$year]))
			$this->initHolidays($region, $type, $year);
		$this->holidayDates[$region][$type][$year][$name] = $this->E($year, $month, $day);
	} /* }}} */

	function holidays_de($year, $type) { /* {{{ */

		switch ($type) {
		case 'public':
			$list = array();
			$list['Neujahrstag']               = $this->E ($year,  1,  1);
			$list['Tag der Arbeit']            = $this->E ($year,  5,  1);
			$list['Tag der Deutschen Einheit'] = $this->E ($year, 10,  3);
			$list['1. Weihnachtstag']          = $this->E ($year, 12, 25);
			$list['2. Weihnachtstag']          = $this->E ($year, 12, 26);

			$easterDate = $this->evaluateMovable ($year, 'easter');

			if ($easterDate) {
				$list['Karfreitag']          = $easterDate - 2;
				$list['Ostermontag']         = $easterDate + 1;
				$list['Christi Himmelfahrt'] = $easterDate +39;
				$list['Pfingstmontag']       = $easterDate +50;
			}
			return $list;
		}
		return null;
	} /* }}} */

	function holidays_at($year, $type) { /* {{{ */

		switch ($type) {
		case 'public':
			$list = array();
			$list['Neujahr']                   = $this->E ($year,  1,  1);
			$list['Heilige Drei Könige']       = $this->E ($year,  1,  6);
			$list['Staatsfeiertag']            = $this->E ($year,  5,  1);
			$list['Mariä Himmelfahrt']         = $this->E ($year,  8, 15);
			$list['Nationalfeiertag']          = $this->E ($year, 10, 26);
			$list['Allerheiligen']             = $this->E ($year, 11,  1);
			$list['Mariä Empängnis']           = $this->E ($year, 12,  8);
			//$list['Heiliger Abend']            = $this->E ($year, 12, 24);
			$list['Christtag']                 = $this->E ($year, 12, 25);
			$list['Stephanitag']               = $this->E ($year, 12, 26);
			//$list['Silvester']                 = $this->E ($year, 12, 31);

			$easterDate = $this->evaluateMovable ($year, 'easter');

			if ($easterDate) {
				//$list['Karfreitag']          = $easterDate - 2;
				$list['Ostermontag']         = $easterDate + 1;
				$list['Christi Himmelfahrt'] = $easterDate +39;
				$list['Pfingstmontag']       = $easterDate +50;
				$list['Fronleichnam']        = $easterDate +60;
			}
			return $list;
		}
		return null;
	} /* }}} */

	//============================================================================
	//
	//	TIMES
	//
	//============================================================================

	//----------------------------------------------------------------------------
	//	true:	match
	//	false:	no match
	//	null:	unknown
	//	default is 00:00-24:00 and Error message
	//----------------------------------------------------------------------------

	function evaluateTimes($request, $acceptEmpty) { /* {{{ */

		$request->_times = array();

		if(!isset($request->tokens[0]['t'])) {
			if (!$acceptEmpty) return "";
			$this->warning ($request, 'No time specification, assuming 0:00-24:00');
			return $request->H < 24;
		} else {
			switch ($request->tokens[0]['t']) {
			case 't':
			case 'e':
				break;
			case '-':
				if ($request->tokens[1]['t']=='t' || $request->tokens[1]['t']=='e') break;
			default:
				if (!$acceptEmpty) return "";
				$this->warning ($request, 'No time specification, assuming 0:00-24:00');
			case '??':
				return $request->H < 24;
			}
		}

		$time = $request->H * 60 + $request->M;
		$result = false;

		for (;;) {

			//-------------------------------------
			//	[time]
			//-------------------------------------

			$first=-1;

			if (!isset($request->tokens[0]['t']) || $request->tokens[0]['t']!='-') {
				$first = $this->evaluateTime ($request);
				if (gettype($first)=='string') return $first;
			}

			$last=$first+1;
			$step=0;

			if(isset($request->tokens[0]['t'])) {
				switch ($request->tokens[0]['t']) {
				//-------------------------------------
				//	time - time
				//-------------------------------------
				case '-':
					array_shift($request->tokens);
					$last = $this->evaluateTime ($request);
					if (gettype($last)=='string') return $last;
					if ($first<0) {
						$h = $last>6*60 ? 6 : 0;
						$first = $h*60;
						$this->error ($request, 'Missing starttime for timerange, assuming '.$h.':00');
					}
					//-------------------------------------
					//	/<step> (for collection_times)
					//-------------------------------------
					if ($this->C($request->tokens, '/')) {
						switch ($request->tokens[0]['t']) {
						case 't':
							$step=$this->G($request->tokens);
							break;
						case 'n':
							$step=$this->G($request->tokens);
							break;
						default:
							return 'H:MM or number of minutes expected';
						}
					} else if ($this->C($request->tokens,'+')) {
						$this->warning ($request, 'Assuming '.
							floor($last/60).':'.($last%60<10?'0':'').($last%60).
							' as endtime for "+" notation');
					}
					break;
				//-------------------------------------
				//	time+
				//-------------------------------------
				case '+':
					array_shift($request->tokens);
					$last = $first + 4 * 60;
					if ($last>=24*60) $last=24*60;
					$this->warning ($request, 'Assuming '.
						floor($last/60).':'.($last%60<10?'0':'').($last%60).
						' as endtime for "+" notation');
					break;
				}
			}

			//-------------------------------------
			//	eval
			//-------------------------------------

			if ($first!=null && $last!=null) {
				if ($last<=$first) $last += 24*60;
				if ($first <= $time && $time < $last) $result = true;
				$request->_times[] = array('f' => $first, 'l' => $step?$last:$first, 's' => $step);
			} else if (!$result) {
				$result = null;
			}

			//-------------------------------------
			//	…, nexttimerange
			//-------------------------------------

			if (isset($request->tokens[0]['t']) && $request->tokens[0]['t'] == ',' && ($request->tokens[1]['t'] == 't' || $request->tokens[1]['t'] == 'e')) {
				array_shift($request->tokens);
				continue;
			}

			//-------------------------------------------------
			//	error correction
			//-------------------------------------------------

			if(isset($request->tokens[0]['t'])) {
				switch ($request->tokens[0]['t']) {
				case 't':	// timespan:	time
				case 'e':	// timespan:	event
					$msg = 'Comma expected before "'.$request->tokens[0].s.'" in "'.$request->rule.'"';
					$this->error ($request, 'Comma expected');
					continue;
				}
			}

			break;
		}

		return $result;
	} /* }}} */

	//----------------------------------------------------------------------------
	//	string:	error
	//	number:	minutes since midnight
	//----------------------------------------------------------------------------

	function evaluateTime($request, $last=null) { /* {{{ */

		//--------------------------------------------
		//	hh:mm
		//--------------------------------------------

		if (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='t') return $this->G($request->tokens);

		//--------------------------------------------
		//	event
		//--------------------------------------------

		if (isset($request->tokens[0]['t']) && $request->tokens[0]['t']=='e') {

			$t = $this->evaluateEvent ($request, $this->G($request->tokens));

			if (gettype($t)=='string') return $t;

			if ($request->tokens[0]['t']=='+' && $request->tokens[1]['t']=='t' && $request->tokens[2]['t']=='hours') {
				$t += $request->tokens[1]['v'];
				array_splice($request->tokens, 0, 3);
				return $t;
			}

			if ($request->tokens[0]['t']=='-' && $request->tokens[1]['t']=='t' && $request->tokens[2]['t']=='hours') {
				$t -= $request->tokens[1]['v'];
				array_splice($request->tokens, 0, 3);
				return $t;
			}

			return $t;
		}

		//--------------------------------------------
		//	syntax error
		//--------------------------------------------

		return 'HH:MM or sunset/sunrise expected';
	} /* }}} */

	//----------------------------------------------------------------------------
	//	number:	minutes from midnight
	//	null:	unknown
	//----------------------------------------------------------------------------

	function evaluateEvent($request, $name) { /* {{{ */

		switch ($name) {

		case 'sunrise':
			$this->warning ($request, 'Using 06:00 for "sunrise"');
			return 6*60+00;

		case 'sunset':
			$this->warning ($request, 'Using 18:00 for "sunset"');
			return 18*60;
		}

		$this->error ($request, 'No time for event "'.$name.'"');
		return null;
	} /* }}} */


	//==============================================================================
	//
	//	Lexical analysis
	//
	//==============================================================================

	//----------------------------------------------------------------------------
	//	scans an expression
	//		result to $request->$request->tokens
	//----------------------------------------------------------------------------

	function scan($request, $expression) { /* {{{ */

		$request->parsedExpression = array();

		$conflicts = array();
		if(preg_match_all ('/[a-z_àè]+[0-9][a-z0-9_]*|[0-9]+[a-z_]\w*/i', $expression, $conflicts))
//		if ($conflicts)
			$this->scanError ($request, 'Missing space at '.implode(', ', $conflicts));

		if(preg_match_all ('/[0-9]+[.h][0-9]+/', $expression, $conflicts))
//		if ($conflicts)
			$this->scanError ($request, 'Please use ":" as hour/minute-separator at '.implode(', ', $conflicts));

		$symbols = array();
		preg_match_all ('/[a-z_]+|24\/7|[0-9]+[.:][0-9]+|[0-9]+[.]|[0-9]+|"[^";]*"|\|\||[^\w\s]/i', $expression, $symbols);

//		sym:
		foreach ($symbols[0] as $f) {

			//------------------------------------------------------------
			//	Integers...
			//	0..60		number
			//	1970..2069	year
			//------------------------------------------------------------

			$v = self::parseInt ($f, 10);
			$l = strtolower($f);

			/* Check for all kinds of numbers in dates and times */
			if (!self::isNaN($v)) {
//				$l2 = $f.split($f.match('/[.]/') ? '.' : ':')[1];
				if(strstr($f, '.') !== false)
					$sepchar = '.';
				else
					$sepchar = ':';
				$l2arr = explode($sepchar, $f);
				if(isset($l2arr[1])) {
					$l2 = $l2arr[1];
					$v2 = self::parseInt ($l2);
				} else {
					$v2 = null;
				}
				if (!self::isNaN($v2)) {
					if (strlen($f)>5 || strlen($l2)!=2 || $v>36 || $v2>=60)
					$this->scanError ($request,
						'Invalid time at "'.$f.'" in rule "'.$request->rule.'"');
					$request->parsedExpression[] = array('t' => 't', 'v' => $v*60+$v2, 's' => $f);
					continue;
				}
				if (count(explode('.', $f))>=2) {
					if ($v<1 || $v>31)
					$this->scanError ($request,
						'Invalid day at "'.$f.'" in rule "'.$request->rule.'"');
					$request->parsedExpression[] = array ('t' => 'd', 'v' => $v, 's' => $f);
					continue;
				}
				if ($f=='24/7') {
					$request->parsedExpression[] = array('t' => $f, 's' => $f);
					continue;
				}
				if ($v<=365) {
					$request->parsedExpression[] = array('t' => 'n', 'v' => $v, 's' => $f);
					continue;
				}
				if ($v>=1970 && $v<2070) {
					$request->parsedExpression[] = array('t' => 'y', 'v' => $v, 's' => $f);
					continue;
				}
			}

			//------------------------------------------------------------
			// wrong words
			// check if symbol is in the list of wrong words
			//------------------------------------------------------------

			foreach (self::$wrong_words as $msg=>$tmp) {
				if(!isset($tmp[$l]))
					continue;
				$ok = $tmp[$l]; 
				$l=strtolower($ok);
				str_replace('<ko>', $f, $msg);
				str_replace('<ok>', $ok, $msg);
				$this->scanError ($request, $msg);
				if ($l=='') continue sym ;
				break;
			}

			//------------------------------------------------------------
			//	Symbols
			//------------------------------------------------------------

			switch ($l) {
			case '+':
			case ',':
			case '-':
			case '/':
			case ':':
			case '[':
			case ']':
				$request->parsedExpression[] = array('t'=>$l, 's'=>$f);
				continue 2;
			case '–':
				$this->scanError ($request,
					'Use a minus sign "-" instead of a dash "–".');
				$request->parsedExpression[] = array('t' => '-', 's' => $f);
				continue 2;
			case ';':
				$request->parsedExpression[] = array('s' => $f);
				continue 2;
			case '||':
				$request->parsedExpression[] = array('s' => $f);
				continue 2;
			}

			//------------------------------------------------------------
			//	Words...
			//	- name of month
			//	- name of weekday
			//	- name of holiday
			//	- other
			//------------------------------------------------------------

			if (isset(self::$seasons[$l])) {
				$v=self::$seasons[$l];
				$request->parsedExpression[] = array('t' => 's', 'v' => $v, 's' => $f);
				continue;
			}
			if (isset(self::$months[$l])) {
				$v=self::$months[$l];
				$request->parsedExpression[] = array('t' => 'm', 'v' => $v, 's' => $f);
				continue;
			}
			if (isset(self::$weekdays[$l])) {
				$v=self::$weekdays[$l];
				$request->parsedExpression[] = array('t' => 'w', 'v' => $v, 's' => $f);
				continue;
			}
			if (isset(self::$movables[$l])) {
				$v=self::$movables[$l];
				$request->parsedExpression[] = array('t' => 'v', 'v' => $v, 's' => $f);
				continue;
			}
			if (isset(self::$holidays[$l])) {
				$v=self::$holidays[$l];
				$request->parsedExpression[] = array('t' => 'h', 'v' => $v, 's' => $f);
				continue;
			}
			if (isset(self::$events[$l])) {
				$v=self::$events[$l];
				$request->parsedExpression[] = array('t' => 'e', 'v' => $v, 's' => $f);
				continue;
			}

			switch ($l) {
			case 'closed':
			case 'days':
			case 'hours':
			case 'off':
			case 'open':
			case 'unknown':
			case 'week':
				$request->parsedExpression[] = array('t' => $l, 's' => $f);
				continue 2;
			}

			//------------------------------------------------------------
			//	Comment
			//------------------------------------------------------------

			if (substr($f,0,1)=='"') {
				$request->parsedExpression[] = array('t'=>'"', 'v'=>$this->T(substr($f,1,-1)), 's'=>$f);
				continue;
			}

			//------------------------------------------------------------
			//	Error correction
			//------------------------------------------------------------

			foreach (self::$wrong_weekdays as $msg=>$tmp) {
				if (!isset($tmp[$l])) continue;
				$v = $tmp[$l];
				$kk = array('Su','Mo','Tu','We','Th','Fr','Sa');
				$ok = $kk[$v];
				$msg = str_replace('<ko>',$f, $msg);
				$msg = str_replace('<ok>',$ok, $msg);
				$this->scanError ($request, $msg);
				$request->parsedExpression[] = array('t'=>'w', 'v'=>$v, 's'=>$f);
				continue 2;
			}

			foreach (self::$wrong_months as $msg=>$tmp) {
				if (!isset($tmp[$l])) continue;
				$v = $tmp[$l];
				$kk = array(null,'Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');
				$ok = $kk[$v];
				$msg = str_replace('<ko>',$f, $msg);
				$msg = str_replace('<ok>',$ok, $msg);
				$this->scanError ($request, $msg);
				$request->parsedExpression[] = array('t'=>'m', 'v'=>$v, 's'=>$f);
				continue 2;
			}

			foreach (self::$wrong_seasons as $msg=>$tmp) {
				if (!isset($tmp[$l])) continue;
				$v = $tmp[$l];
				$kk = array(null,'Spring','Summer','Autumn','Winter');
				$ok = $kk[$v];
				$msg = str_replace('<ko>',$f, $msg);
				$msg = str_replace('<ok>',$ok, $msg);
				$this->scanError ($request, $msg);
				$request->parsedExpression[] = array('t'=>'s', 'v'=>$v, 's'=>$f);
				continue 2;
			}

			//------------------------------------------------------------
			//	Symbols
			//------------------------------------------------------------

			$request->parsedExpression[] = array('t'=>'??', 's'=>$f);
		}

		//------------------------------------------------------------
		//	eof marker
		//------------------------------------------------------------

		$request->parsedExpression[] = array();
	} /* }}} */

	//----------------------------------------------------------------------------
	//	little helper functions
	//----------------------------------------------------------------------------

	function T($s) { /* {{{ */
		return trim($s);
//FIXME		return $s.toString().replace (/^\s+/, '').replace (/\s+$/, '');
	} /* }}} */

	/**
	 * Get the value of the current token and remove the token from the list
	 *
	 * @param array $tokens list of remaining tokens
	 * @return string token value
	 */
	function G(&$tokens) { /* {{{ */
		$v = $tokens[0]['v'];
		array_shift($tokens);
		return $v;
	} /* }}} */

	/**
	 * Check if the current token has the given tag and return its value
	 * if the name matches
	 * If it has the tag, then the token will be shifted from the list of tokens
	 *
	 * @param array $tokens list of remaining tokens
	 * @param string $tag tag name of token to check for
	 * @return string token value if current token has given tag, otherwise null
	 */
	function CG(&$tokens, $tag) { /* {{{ */
		if (!isset($tokens[0]['t']) || $tokens[0]['t']!=$tag) return null;
		$v = $tokens[0]['v'];
		array_shift($tokens);
		return $v;
	} /* }}} */

	/**
	 * Check if the current token has the given tag
	 * If it has the tag, then the token will be shifted from the list of tokens
	 *
	 * @param array $tokens list of remaining tokens
	 * @param string $tag tag name of token to check for
	 * @return boolean true if current token has given tag, otherwise false
	 */
	function C(&$tokens, $tag) { /* {{{ */
		if (!isset($tokens[0]['t']) || $tokens[0]['t']!=$tag) return false;
		array_shift($tokens);
		return true;
	} /* }}} */

	//============================================================================
	//
	//	DATE / TIME
	//
	//============================================================================

	//----------------------------------------------------------------------------
	//	create timestamp from Data()-object or component-object
	//----------------------------------------------------------------------------

	function createTimestamp($timestamp, $offset=null, $shift=0) { /* {{{ */

		$result = new OSM_OpeningHours_Request();
		$now    = new OSM_OpeningHours_Date();

		if ($timestamp==null) $timestamp = $now;

		if (is_object($timestamp) && get_class($timestamp) == 'OSM_OpeningHours_Date') {
			//--------------------------------------------
			//	components from JS Date-object
			//--------------------------------------------
			$result->S = $timestamp->getSeconds();
			$result->M = $timestamp->getMinutes();
			$result->H = $timestamp->getHours();
			$result->d = $timestamp->getDate();
			$result->m = $timestamp->getMonth();
			$result->y = $timestamp->getFullYear();
		} else {
			//--------------------------------------------
			//	parse components from input argument
			//--------------------------------------------
			$result->S = self::parseInt($timestamp->S,10); if (self::isNaN($result->S)) $result->S=0;
			$result->M = self::parseInt($timestamp->M,10); if (self::isNaN($result->M)) $result->M=0;
			$result->H = self::parseInt($timestamp->H,10); if (self::isNaN($result->H)) $result->H=0;
			$result->d = self::parseInt($timestamp->d,10); if (self::isNaN($result->d)) $result->d=1;
			$result->m = self::parseInt($timestamp->m,10); if (self::isNaN($result->m)) $result->m=1;
			$result->y = self::isNaN (self::parseInt ($timestamp->y, 10)) ?
				$now->getFullYear() : (self::parseInt ($timestamp->y, 10)+30)%100+1970;
		}

		//-----------------------------------------------------------
		//	offset is a real offset to the input argument time
		//-----------------------------------------------------------

		if ($offset) $result->S += $offset;

		//-----------------------------------------------------------
		//	shift forces the hour to be >= 24
		//	used to implement day wrap
		//-----------------------------------------------------------

		if ($shift ) $result->H -= $shift;

		//-----------------------------------------------------------
		//	normalize time
		//-----------------------------------------------------------

		$carry = floor ($result->S / 60); $result->S -= $carry * 60; $result->M += $carry;
		$carry = floor ($result->M / 60); $result->M -= $carry * 60; $result->H += $carry;
		$carry = floor ($result->H / 24); $result->H -= $carry * 24; $result->d += $carry;

		//-----------------------------------------------------------
		//	normalize date
		//-----------------------------------------------------------

		if ($result->d<1) {
			$result->m--;
			$result->d=$this->E($result->y,$result->m+1,1)-$this->E($result->y,$result->m,1);
		}
		if ($this->E($result->y,$result->m,$result->d)>=$this->E($result->y,$result->m+1,1)) {
			$result->m++; $result->d=1;
		}
		if ($result->m< 1) {
			$result->y--;
			$result->m=12;
		}
		if ($result->m>12) {
			$result->y++;
			$result->m= 1;
		}

		//-----------------------------------------------------------
		//	shift forces the hour to be >= 24
		//	used to implement day wrap
		//-----------------------------------------------------------

		if ($shift) $result->H += $shift;

		//-----------------------------------------------------------
		//	compute absolute day number and weekday
		//-----------------------------------------------------------

		$result->s = $this->E ($result->y,1,1);
		$result->e = $this->E ($result->y,$result->m,$result->d);
		$result->w = $this->W ($result->e);
		$result->l = $this->L ($result)+1;
		$result->wy= $result->y; $s = $this->S ($result->wy); if ($result->e<$s) $s=$this->S(--$result->wy);
		$result->yw= floor(($result->e-$s)/7)+1;
		return $result;
	} /* }}} */

	//----------------------------------------------------------------------------
	//	little helper functions
	//----------------------------------------------------------------------------

	function E($y,$m,$d) { /* {{{ */
		return 367*$y - floor (1.75 * ($y + floor (($m+9)/12))) + floor(275*$m/9) + $d - 719574;
	} /* }}} */

	function L($request) { /* {{{ */
		return $this->E($request->y, $request->m+1, 1) - $this->E($request->y, $request->m, 1);
	} /* }}} */

	function W($d) { /* {{{ */
		return ($d+4) % 7;
	} /* }}} */

	function S($y) { /* {{{ */
		$s = $this->E ($y, 1, 1);
		return $s + (11-$this->W($s))%7-3;
	} /* }}} */

	function YW($y, $d) { /* {{{ */
		$s = $this->S ($y);
		if ($d<$s) $s = $this->S(--$y);
		return array('y'=>$y, 'w'=>floor(($d-$s)/7)+1);
	} /* }}} */

}

?>
